# Copyright (c) 2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import socket
import random
import json

from flexapiserver.log import *
import flexapiserver.abort
import flexapiserver.response

FLEX_TIMEOUT = 60

class Client:
    def __init__(self, conf):
        __slots__ = ('use_unix_socket', 'sock_path', 'current_server', 'port',
                     'back_socket', 'back_file', 'am_connected', 'conf')

        if 'flex_backend_servers' in conf:
            assert isinstance(conf['flex_backend_servers'], list), "flex_backend_servers is not a list"
            assert len(conf['flex_backend_servers']) > 0, "No flex_backend_servers defined"
            assert int(conf['flex_backend_port']) > 0, "Invalid flex_backend_port"
            self.port = int(conf['flex_backend_port'])
            self.use_unix_socket = False
        elif 'flex_backend_sock_path' in conf:
            self.sock_path = conf['flex_backend_sock_path']
            self.use_unix_socket = True

        self.conf = conf
        self.back_socket = None
        self.back_file = None
        self.am_connected = False

        # 9/30/20 djw: removed the self.connect() as it will be done on a query.
        # if we do it now, then the FlexApiServer UWSGI Master opens a useless connection.
        #self.connect()

    def close(self):
        self.am_connected = False

        try:
            if self.back_file is not None:
                self.back_file.close()
        except Exception as e:
            log(DEBUG, "backend-connection: Problem closing backend file overlay:" + repr(e))

        self.back_file = None

        try:
            if self.back_socket is not None:
                self.back_socket.shutdown(socket.SHUT_RDWR)
                self.back_socket.close()
        except Exception as e:
            log(DEBUG, "backend-connection: Problem closing backend connection:" + repr(e))

        self.back_socket = None

    def connect(self):
        ''' Return True if connected, else False
        '''
        if self.am_connected:
            self.close()

        if self.use_unix_socket:
            self.back_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            if self.back_socket is None:
                log(DEBUG, "backend-connection: socket.socket failed")
                return False

            # too noisy log(DEBUG, "backend-connection: Connecting to flex-backend on unix socket {}".format(self.sock_path))
            try:
                self.back_socket.connect(self.sock_path)
            except socket.timeout:
                log(DEBUG, "backend-connection: Connect timeout to flex-backend {}".format(self.sock_path))
                self.close()
                # will have to retry later
                return False
            except ConnectionRefusedError:
                log(DEBUG, "backend-connection: Connect refused to flex-backend {}".format(self.sock_path))
                self.close()
                # will have to retry later
                return False
            except Exception as e:  # Possibly FileNotFound
                log(DEBUG, "backend-connection: Other Connect failure to flex-backend {}".format(self.sock_path) + ": " + str(e))
                self.close()
                # will have to retry later
                return False
        else:
            self.back_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            if self.back_socket is None:
                log(DEBUG, "backend-connection: socket.socket failed")
                return False

            self.current_server = random.choice(self.conf['flex_backend_servers'])

            log(DEBUG, "backend-connection: Connecting to flex-backend {} on TCP port {}".format(self.current_server, self.port))
            try:
                self.back_socket.connect((self.current_server, self.port))     # blocking call
            except socket.timeout:
                log(DEBUG, "backend-connection: Connect timeout to flex-backend {} on TCP port {}".format(self.current_server, self.port))
                self.close()
                # will have to retry later
                return False
            except ConnectionRefusedError:
                log(DEBUG, "backend-connection: Connect refused to flex-backend {} on TCP port {}".format(self.current_server, self.port))
                self.close()
                # will have to retry later
                return False

        self.back_file = self.back_socket.makefile(mode='r', buffering=65336)

        self.am_connected = True

        return True

    def _send_str(self, msg):
        msg = (msg+'\n').encode("utf-8")

        try:
            self.back_socket.sendall(msg) # only returns None or raises exception
            return True
        except socket.timeout:
            log(DEBUG, "backend-connection: sendall timeout writing")
            self.close()
            return False
        except Exception as e:
            log(DEBUG, "backend-connection: sendall failed with " + str(e))
            self.close()
            return False

    def send_query(self, q_json):
        '''
        Query object: {"query":{"search_how":"...",...}}
        This takes a json object that is inside the "query".
        Returns True if sent, otherwise False
        '''
        if not self.connect():
            log(DEBUG, "backend-connection: send_query -- could not connect so not sending")
            return False

        str = '{"query":'+ json.dumps(q_json, separators=(',', ':')) + '}'
        self._send_str(str)
        return True

    def results_iterator(self):
        ''' Iterate over all results.  Raise InternalServerError if connection fails
        '''
        if not self.am_connected:
            log(DEBUG, "backend-connection: in results_iterator but not connected")
            raise flexapiserver.response.InternalServerError

        while True:
            try:
                line = self.back_file.readline()
            except Exception as ex:
                log(DEBUG, "backend-connection: readline failed with " + str(ex))
                self.close()
                raise ex
            if not line:
                log(DEBUG, "readline returned None")
                self.close()
                raise flexapiserver.response.InternalServerError
            line = line.rstrip('\n')
            yield line

