# Copyright (c) 2012-2016, 2018-2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import collections
import json
import os
import ipaddress

CONF_FILE = os.environ.get('DNSTLI_CONFIG_FILE', '/etc/dnstli/flexapiserver.conf')
API_DIR = '/etc/dnstli/auth-api'

# for minimum and maximum version number checks
MY_API_VERSION = 2

DEFAULT_MAX_QUERY_TIME = 999999     # very long number of seconds
DEFAULT_OFFSET_MAX = 999999999999   # very large number of rows

required_config_keys = (
    'res_limit_default',
    'res_limit_max',
)

optional_config_keys = (
    'flex_backend_sock_path',
    'flex_backend_servers',
    'flex_backend_port',
    'export_installation', # true if an export installation
    'admin_addresses', # list of addresses that may connect as any account
    'query_limit_default',
    'query_limit_period',
    'query_limit_redis_socket',
    'query_limit_redis_url',
    'max_query_time',
    'offset_max',
    'cors_origins',
)

positive_integer_config_keys = (
    'res_limit_default',
    'res_limit_max',
    'offset_max',
    'query_limit_default',
    'query_limit_period',
    'flex_backend_port'
)

nullable_positive_float_config_keys = (
    'max_query_time',
)

def get_config(config_fname=None):
    '''
    Returns a configuration dictionary, with the above required_config_keys and possibly optional_config_keys.
    Raises exceptions with any problems it finds in the config.
    '''
    if config_fname is None:
        config_fname = CONF_FILE

    with open(config_fname) as f:
        print("loading config from file {}\n".format(config_fname))
        try:
            config = json.load(f)
        except Exception as e:
            raise Exception("loading config from file {} failed with {}\n".format(config_fname,str(e)))

    for k in required_config_keys:
        if not k in config:
            raise Exception('config key "%s" is required but not present in config file %s' % (k, config_fname))

    for k in config:
        if not (k in required_config_keys or k in optional_config_keys):
            raise Exception('config key "%s" is unknown (and thus unexpected) in config file %s' % (k, config_fname))

    for k in positive_integer_config_keys:
        if k in config:
            try:
                config[k] = int(config[k])
            except ValueError:
                raise Exception('config key "%s" is not an integer in config file %s' % (k, config_fname))
            if config[k] < 0:
                raise Exception('config key "%s" is a negative integer in config file %s' % (k, config_fname))

    for k in nullable_positive_float_config_keys:
        if k in config:
            try:
                config[k] = float(config[k])
            except ValueError:
                raise Exception('config key "%s" is not a float in config file %s' % (k, config_fname))
            if config[k] < 0:
                raise Exception('config key "%s" is a negative float in config file %s' % (k, config_fname))

    if not isinstance(config.get('flex_backend_servers', []), collections.Iterable):
        raise Exception('flex_backend_servers must be iterable in config file %s' % config_fname)

    if 'flex_backend_servers' in config and 'flex_backend_sock_path' in config:
        raise Exception('flex_backend_servers and flex_backend_sock_path are mutually exclusive in config file %s' % config_fname)

    # if these are missing, then use defaults
    if config.get('max_query_time') == None:
        config['max_query_time'] = DEFAULT_MAX_QUERY_TIME
    if config.get('offset_max') == None:
        config['offset_max'] = DEFAULT_OFFSET_MAX

    if 'query_limit_redis_url' in config and 'query_limit_redis_socket' in config:
        raise Exception('query_limit_redis_url and query_limit_redis_socket are mutually exclusive in config file %s' % config_fname)

    if 'query_limit_redis_url' in config and not (
            'query_limit_default' in config and
            'query_limit_period' in config):
        raise Exception('setting query_limit_redis_url requires also setting query_limit_default, query_limit_period in config file %s' % config_fname)

    if 'query_limit_redis_socket' in config and not (
            'query_limit_default' in config and
            'query_limit_period' in config):
        raise Exception('setting query_limit_redis_socket requires also setting query_limit_default, query_limit_period in config file %s' % config_fname)

    if not isinstance(config.get('cors_origins', []), collections.Iterable):
        raise Exception('cors_origins must be iterable in config file %s' % config_fname)

    if not isinstance(config.get('admin_addresses', []), collections.Iterable):
        raise Exception('admin_addressess must be iterable in config file %s' % config_fname)

    for acl in config.get('admin_addresses', []):
        try:
            # if acl is not expressed as a network, it will be treated as a
            # /32 (IPv4) or /128 (IPv6).  This will fail if expressed as a
            # network but host bits are set.
            dummy = ipaddress.ip_network(acl)
        except ValueError as e:
                raise Exception('Invalid admin acl %s: %s in config file %s' % (acl, e, config_fname))

    if 'export_installation' in config: # optional config parameter but if present, must be a bool
        ei = config.get('export_installation')
        if not isinstance(ei, bool):
            raise Exception("'export_installation' value %s is not a boolean (True or False) in config file %s"
                            % (ei, config_fname))

        if ei and ('query_limit_redis_socket' in config or 'query_limit_redis_url' in config):
            raise Exception("An export_installation may not have redis config parameters defined, in config file %s"
                            % (config_fname))

    return config
