# Copyright (c) 2019-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import datetime
import json

from flexapiserver.log import *

class HTTPResponse(Exception):
    __slots__ = ('status', 'headers', 'message',)
    def __init__(self, status, message):
        self.status = status
        if message == None:
            self.message = None
        else:
            self.message = (message + '\n').encode('utf-8')
        self.headers = [('Content-Type', 'text/plain')]

Raw200Status = '200 OK'
Raw400Status = '400 Bad Request'
Raw401Status = '401 Unauthorized'
Raw403Status = '403 Forbidden'
Raw415Status = '415 Unsupported Media Type'
Raw416Status = '416 Requested Range Not Satisfiable'
Raw429Status = '429 Too Many Requests'
Raw500Status = '500 Internal Server Error'

BadRequest = HTTPResponse(
    Raw400Status, 'Error: unable to parse request')

RegExSyntaxError = HTTPResponse(
    Raw400Status, 'Error: Regular expression has a syntax error')
RegExUnsupported = HTTPResponse(
    Raw400Status, 'Error: Regular expression uses an unsupported feature')
GlobSyntaxError = HTTPResponse(
    Raw400Status, 'Error: Glob has a syntax error')
GlobUnsupported = HTTPResponse(
    Raw400Status, 'Error: Glob uses an unsupported feature')

QuotaExpired = HTTPResponse(
    Raw401Status, 'Error: Quota is expired')
APIVersionMismatch  = HTTPResponse(
    Raw401Status, 'Error: Your API key may not query this API version')

MissingAPIKey = HTTPResponse(
    Raw403Status, 'Error: X-API-Key header not present')
BadAPIKey = HTTPResponse(
    Raw403Status, 'Error: Bad API key')
RemoteAddressUnauthorized = HTTPResponse(
    Raw403Status, 'Client IP address not authorized for account')

UnsupportedMediaType = HTTPResponse(
    Raw415Status, 'Error: The Accept: header does not specify a supported content type for this query')

OffsetOverMax = HTTPResponse(
    Raw416Status, 'Error: offset value greater than maximum allowed')
OffsetDisallowed = HTTPResponse(
    Raw416Status, 'Error: offset not allowed for this type of API key')
OffsetUnsupported = HTTPResponse(
    Raw416Status, 'Error: offset not allowed for this type of query')

RateLimitExceeded = HTTPResponse(
    Raw429Status, 'Error: Rate limit exceeded')
BurstRateExceeded = HTTPResponse(
    Raw429Status, 'Error: Burst window rate limit exceeded')

InternalServerError = HTTPResponse(
    Raw500Status, 'Error: exception occurred')
BackendServerError = HTTPResponse(
    Raw500Status, 'Error: Backend server is not responding')

def customBadRequest(msg):
    ''' for other errors, with details
    '''
    return HTTPResponse(Raw400Status, 'Error: {}'.format(msg))

cond_begin = 'begin'
cond_ongoing = 'ongoing'
cond_succeeded = 'succeeded'
cond_limited = 'limited'
cond_failed = 'failed'

def format_saf_row(cond = None, obj = None, msg = None):
    row = '{'
    if cond:
        row += '"cond":"' + cond + '"'
        if obj or msg:
            row += ','
    if obj:
        row += '"obj":' + obj
        if msg:
            row += ','
    if msg:
        # backslash escape any backslash escapes, etc.
        row += '"msg":' + json.dumps(msg)

    row += '}\n'

    if cond == cond_limited or cond == cond_failed:
        log(DEBUG, "format_saf_row cond=" + str(cond) + " msg=" + msg)

    return row

def format_http500(logid):
    return '''
<html>
<head><title>500 Internal Server Error</title></head>
<body>
<p>An internal server error has occurred. The log ID for this error is: {}.</p>
<p>Sorry about that.</p>
</body>
</html>
'''.format(logid).strip()

def fmt_time(epoch, use_epoch_time):
    ''' format times '''
    if use_epoch_time:
        return str(epoch)
    else:
        return '"' + datetime.datetime.utcfromtimestamp(epoch).isoformat("T")+'Z"'

def content_type_to_return(env, allowed_cts, preferred_ct):
    '''
    Check if the HTTP_ACCEPT header in env contains one of the allowed content types.
    Return the first matching one.
    If find the wild card */* then use the preferred content type.
    Raises UnsupportedMediaType if none of the content types are allowed.
    '''
    if 'HTTP_ACCEPT' in env:
        # extract content types acceptable from HTTP request
        ctypes = env['HTTP_ACCEPT'].split(',')
        # for each, lower case it and strip out any priority field
        for ctplus in ctypes:
            ct = ctplus.strip().lower().split(';', 1)[0]
            if ct == "*/*":
                return preferred_ct
            if ct in allowed_cts:
                return ct
        log(DEBUG, "Saw only unsupported values in HTTP_ACCEPT "  + str(ctypes))
    else:
        log(DEBUG, "Request does not include a HTTP_ACCEPT header")
    raise UnsupportedMediaType

