# Copyright (c) 2012-2016, 2018-2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import binascii
from collections import OrderedDict
import io
import itertools
import json
import time
import traceback
import uuid
import urllib.parse
import re
import sys

import wdns

from flexapiserver.log import *
import flexapiserver.abort
import flexapiserver.conf
import flexapiserver.filter
import flexapiserver.response
import flexapiserver.util

conf = flexapiserver.conf.get_config()

failed_incomplete = flexapiserver.response.format_saf_row(
    cond = 'failed', msg = "Processing timeout; results may be incomplete").encode('utf-8')

failed_internal = flexapiserver.response.format_saf_row(
    cond = 'failed', msg = "Internal error").encode('utf-8')

heartbeat_saf = flexapiserver.response.format_saf_row() # empty

# Truncate swclient, version, and id parameters for logging purposes.
MAX_SWCLIENT_LENGTH = 20
MAX_VERSION_LENGTH = 20
MAX_ID_LENGTH = 30

# Flexible Internal API Fin map -- fin codes to meanings
flexIntApiFinMap = {
#   200: 'success',                 # never separately reported
    400: 'Bad request',             # common
#   404: 'no answers',              # handled specially below
    406: 'Illegal param type',      # NOT USED in flex-backend currently
    416: 'Bad param value',         # NOT USED in flex-backend currently
    500: 'Internal error',
    504: 'Gateway timeout',
    801: 'Regex syntax error',
    802: 'Regex syntax unsupported',
    803: 'Glob syntax error',
    804: 'Glob syntax unsupported', # NOT USED in flex-backend currently
    805: 'Bad search method',
    806: 'Bad search what parameter',
    807: 'Bad return what parameter',
    808: 'Bad rrtype',
    809: 'Illegal limit',
    810: 'Search string unsupported',
    811: 'Bad exclude expression'
};

def mapFin(fin):
    ''' Lookup a fin code in the map for its english description.
        If not found, return a synthetic msg
    '''
    return flexIntApiFinMap.get(fin, "Internal error #"+ str(fin))

def log_lookup_error(env, status, search_method, tb=None, logid=None):
    msg = flexapiserver.util.form_basic_log(env, search_method + '-error')
    msg['status'] = status
    if logid:
        msg['logid'] = logid
    if tb:
        msg['traceback'] = json.dumps(tb) # escape embedded quotes etc
    msg['lookup'] = { 'server': 'flex', 'error': status}

    log(NOTICE, json.dumps(msg))

def extractBoolean(qs, flagName):
    ''' Return True/False for a flag.  Return False if unclear or not specified.
        Actually returns False unless the string starts with t or T.
    '''
    try:
        flag = qs.get(flagName, None)
        if flag:
            flag = flag[0]
            if len(flag) > 0 and (flag[0] == 't' or flag[0] == 'T'):
                return True
    except:
        pass
    return False

def extractString(qs, flagName):
    ''' Return a string extracted from the query string. Return None if not specified.
    '''
    try:
        strList = qs.get(flagName, None)
        if str:
            return strList[0]
    except:
        pass
    return None

def extractLimit(env, qs):
    ''' Return limit or return 0 if it is missing.  Raise BadRequest if anything is wrong.

        There are four controls for the limit on the number of results to return.  In ascending priority:
        - global dnstli config default, res_limit_default, used if query doesn't specify a limit and is not reduced
        - limit query parameter, limit, may be reduced by next two
        - global dnstli config max, res_limit_max, which will also be used if query limit parameter is 0
        - per-subscription configuration, flexapiserver.res_limit_override
          res_limit_override is an optional per-user config element
    '''
    try:
        limit = int(qs['limit'][0])
        if limit < 0:
            raise flexapiserver.response.BadRequest
        try:
            if limit > conf['res_limit_max'] or limit == 0:
                limit = conf['res_limit_max']
        except ValueError:
            raise flexapiserver.response.BadRequest
        except:
            limit = conf['res_limit_default']

        return limit
    except ValueError:
        raise flexapiserver.response.BadRequest
    except KeyError:
        return conf['res_limit_default']   # use the default if limit was not given

def extractId(qs):
    ''' Return a well-formed user id or return None if anything is wrong or if it is missing '''
    id = qs.get('id')
    if id:
        return id[0][:MAX_ID_LENGTH]
    return None

def extractOffset(qs):
    ''' Return integer number of rows to offset (e.g. skip) if the query string requests it.  Return 0 if not specified.
    '''
    try:
        offset = qs.get('offset', 0)[0]
        if offset:
            offset = int(offset)
            if offset > 0:
                return offset
    except:
        pass
    return 0

def extractSWClient(qs):
    ''' Return a well-formed "swclient" or return None if anything is wrong or if it is missing.
        In APIv1, as a fallback for the previous deployed version of Scout, if "swclient" is missing,
        then tried "client".  This fallback is not necessary for APIv2.
    '''
    # Try "swclient" first, fallback to "client"
    swclient = qs.get('swclient', None)
    if swclient:
        return swclient[0][:MAX_SWCLIENT_LENGTH]
    return None

def extractVersion(qs):
    ''' Return a well-formed user version or return None if anything is wrong or if it is missing '''
    version = qs.get('version')
    if version:
        return version[0][:MAX_VERSION_LENGTH]
    return None

# The following query parameters are already logged separately.
# We want to log all the others.
alreadyLoggedQueryParams = ('swclient', 'id', 'version')

def extractOtherQueryParamsToLog(qs):
    ''' Return other query parameters from a query string that we do not already log in some fashion.
        Return None if there is any error or there are none to log..

        The other parameters will go into a new keyvalue in the "lookup" keyvalue called
        "otherParams", if there are any of them.

        There are some parameters that won't be logged.
        Given "&param=4&param=2&keyonly" only "param": "4" will be logged.
    '''
    try:
        others = dict()
        for qp in qs:
            if qp not in alreadyLoggedQueryParams:
                others[qp] = qs[qp][0]
        return others
    except:
        return None

# Which content types are allowed
allowed_query_content_types = {
    'application/ldjson',
    'application/x-ldjson',
    'application/ndjson',
    'application/x-ndjson',
    'application/jsonl',
    'application/x-jsonl' }

# The preferred content type if */* is requested
preferred_query_content_type = 'application/x-ndjson'

class QueryTimeExceeded(Exception):
    pass

class QueryTimeExceededIncomplete(Exception):
    pass

class FLEXAPI:
    def __init__(self, client):
        self.client = client

    def __call__(self, env, start_response):
        ''' This is called by auth.py code like: return self.app(env, start_response)
        '''
        search_method = 'flex'    # need this defined for the except clause
        try:
            uri = env.get('REQUEST_URI', None)
            if not uri:
                raise flexapiserver.response.customBadRequest("REQUEST_URI not available from environment")
            #
            # Initial fixed values: /dnsdb/v2
            # Required: search_method, what_to_search, value
            # Optional: rrtype
            # Parameters: limit, time_first_before, time_first_after,
            #     time_last_before, time_last_after, offset, swclient,
            #     version, id, exclude
            #
            # ^dnsdb/v2/(regex|glob)/(rrnames|rdata)/[^/]+/([a-zA-Z0-9])?
            #
            # Therefore, the minimum number of path segments is 5 and the maximum is 6.

            # parse the URL and extract just the path segments from it
            pathSegments = urllib.parse.urlparse(uri).path.strip('/').split('/')

            if len(pathSegments) < 5:
                raise flexapiserver.response.customBadRequest("REQUEST_URI is missing path segment")
            if len(pathSegments) > 6:
                raise flexapiserver.response.customBadRequest("REQUEST_URI has too many path segments")

            if pathSegments[0] != 'dnsdb' and pathSegments[1] != 'v2':
                raise flexapiserver.response.customBadRequest("REQUEST_URI does not start with /dnsdb/v2/")

            search_method = pathSegments[2]
            if search_method != 'regex' and search_method != 'glob':
                raise flexapiserver.response.customBadRequest("SEARCH_METHOD has an unsupported value")

            what_to_search = pathSegments[3]
            if what_to_search != 'rrnames' and what_to_search != 'rdata':
                raise flexapiserver.response.customBadRequest("WHAT_TO_SEARCH has an unsupported value")

            value  = urllib.parse.unquote(pathSegments[4])

            # optional rrtype with default to ANY.  If present, check if allowed for that type of search
            if len(pathSegments) == 6:
                rrtype = pathSegments[5]
                if rrtype.upper() == "ANY-DNSSEC":
                    raise flexapiserver.response.customBadRequest("RRType has an unsupported DNSSEC value")
                try:
                    rrtype_num = wdns.str_to_rrtype(rrtype) # convert just to see if get an exception
                except:
                    raise flexapiserver.response.customBadRequest("RRType has an unsupported value")
                if flexapiserver.util.is_dnssec(rrtype_num):
                    raise flexapiserver.response.customBadRequest("RRType has an unsupported DNSSEC value")
                if what_to_search == 'rdata' and flexapiserver.util.forbidden_rtype_for_rdata(rrtype_num):
                    raise flexapiserver.response.customBadRequest("rdata searches are not allowed for the specified RRType")
            else:
                rrtype = "ANY"

            qs = urllib.parse.parse_qs(env.get('QUERY_STRING',''))

            xLimit = extractLimit(env, qs) # may raise BadRequest
            xId = extractId(qs)
            xOffset = extractOffset(qs)
            xSWClient = extractSWClient(qs)
            xVersion = extractVersion(qs)
            xExclude = extractString(qs, 'exclude')
            if xExclude and len(xExclude) == 0: # if &exclude is empty then do not use it
                xExclude = None

            # note, extracting the following fields only in filter.py:
            # time_first_before, time_first_after, time_last_before, time_last_after

            returnContentType = flexapiserver.response.content_type_to_return(env,
                                                                             allowed_query_content_types,
                                                                             preferred_query_content_type)

            xOtherQueryParamsToLog = extractOtherQueryParamsToLog(qs)

            # survived extracting fields without raising an error -- query is maybe ok

        except flexapiserver.response.HTTPResponse as err:
            log_lookup_error(env, err.status, search_method)
            raise
        except Exception as e:
            log_exception()
            raise flexapiserver.response.InternalServerError

        return Lookup(self.client, env, qs, start_response, returnContentType, search_method, what_to_search,
                      value, rrtype, xLimit, xId, xOffset, xSWClient,
                      xVersion, xExclude, xOtherQueryParamsToLog)

class Lookup:
    '''The uwsgi python interface requires this to return an iterable yielding zero or more bytestrings.  This
    should not assume the entire iterator will be consumed, as it may be closed early by the server.
    '''
    def __init__(self, client, env, qs, start_response, returnContentType, search_method, what_to_search,
                 value, rrtype, limit, id, offset, swclient,
                 version, exclude, otherQueryParamsToLog):
        self.client = client
        self.env = env
        self.qs = qs
        self.start_response = start_response
        self.returnContentType = returnContentType
        self.search_method = search_method
        self.what_to_search = what_to_search
        self.value = value
        self.rrtype = rrtype
        self.limit = limit
        self.id = id
        self.offset = offset
        self.swclient = swclient
        self.version = version
        self.exclude = exclude
        self.otherQueryParamsToLog = otherQueryParamsToLog

        self.max_query_time = conf['max_query_time']
        self.offset_max = conf['offset_max']
        self.row_count = 0
        self.bytes = 0
        self.secs = 0
        self.secs_to_first = None

        self.fin = None
        self.msg = None

    def pager(self, iterator, max_bytes=4096):
        ''' Rebuffer the output from the iterator in large chunks for
            more-efficient networking, except if we see heartbeat
            then output the chunk immediately.
        '''
        buf = io.StringIO()

        for chunk in iterator:
            buf.write(chunk)

            # Note that the heartbeat_saf comparison is just a heuristic.
            # The current value of it is just "{}" which is only one way
            # of expressing a heartbeat.  If we internally used a non-minimal form
            # (i.e. '{"cond":"ongoing"}') then this heuristic would not work.
            if chunk == heartbeat_saf or buf.tell() >= max_bytes:
                yield buf.getvalue()
                buf = io.StringIO()
        else:
            if buf.getvalue():
                yield buf.getvalue() # yield final trailing fragment

    def format_response(self):
        ''' Iterator over the results, converting them to COF format.
        '''

        for entry in self.resultsIterator:
            js = json.loads(entry)
            now = time.time()

            # even if we get heartbeats, still check for query timeouts
            if self.max_query_time and now - self.t0 > self.max_query_time:
                raise QueryTimeExceededIncomplete

            # do not count a heartbeat as the first result yet
            if "heartbeat" in js:
                # only for developer: log(DEBUG, "Heartbeat from flex-backend: " + entry)
                yield heartbeat_saf
                continue

            if self.secs_to_first is None:
                self.secs_to_first = now - self.t0

            if "fin" in js:
                self.fin = js['fin']
                if "msg" in js:
                    self.msg = js['msg']
                self.secs = now - self.t0
                return

            # Two cases to handle: rrnames and rdata.
            # Shuffle the JSON variables around to go from the internal-FLEX-API to external-FLEX-API.

            # Use an OrderedDict to make the final JSON object "pretty" like the API doc.
            saf_obj = OrderedDict()
            wellFormed = True
            if "terse" in js:
                inner = js['terse']
                if self.what_to_search == 'rrnames':
                    if 'rrname' in inner and 'rrtype' in inner:
                        saf_obj['rrname'] = inner['rrname']
                        saf_obj['rrtype'] =  inner['rrtype']
                    else:
                        wellFormed = False
                else:
                    if 'rdata' in inner and 'rrtype' in inner and 'raw_rdata' in inner:
                        saf_obj['rdata'] =  inner['rdata']
                        saf_obj['rrtype'] =  inner['rrtype']
                        saf_obj['raw_rdata'] =  inner['raw_rdata']
                    else:
                        wellFormed = False
            else:
                wellFormed = False

            if not wellFormed:
                log(DEBUG,"Malformed answer from flex-backend '" + entry + "'=" + str(js))
                raise flexapiserver.response.InternalServerError

            saf_obj = json.dumps(saf_obj, separators=(',', ':'))
            self.row_count += 1
            yield flexapiserver.response.format_saf_row(obj = saf_obj)

    def __iter__(self):
        sent_headers = False
        try:
            self.begin_lookup()
            self.start_response(flexapiserver.response.Raw200Status, [('Content-Type', self.returnContentType)])
            sent_headers = True
            yield flexapiserver.response.format_saf_row(flexapiserver.response.cond_begin).encode('utf-8')

            # pager forms large-size chunks of answers containing multiple rows
            for chunk in self.pager(self.format_response()):
                self.bytes += len(chunk)
                yield chunk.encode('utf-8')
                # after every chunk, look if the customer disconnected
                flexapiserver.abort.check_connection()
            if self.fin == 200:
                cond = flexapiserver.response.cond_succeeded
                self.msg = None
            elif self.fin == 404: # no content
                cond = flexapiserver.response.cond_succeeded
                self.msg = "No results found"
            else:
                cond = flexapiserver.response.cond_failed
                if self.msg is None:
                    self.msg = "No details provided"
                self.msg = mapFin(self.fin) + ": " + self.msg
            self.client.close()
            self.log_lookup()
            yield flexapiserver.response.format_saf_row(cond = cond, msg = self.msg).encode('utf-8')
            return
        except flexapiserver.abort.Abort:
            self.client.close()
            # This case is when the backend yielded a Timeout and we checked for a client
            # connection, but it was dead so we raised abort to kill the query procesing.  There is
            # no client attached to which to report the query failure.
            yield failed_incomplete
            self.log_lookup(early_exit=True, early_cause="query_abort")
        except QueryTimeExceeded:
            self.client.close()
            # This case is when above in __iter__, we detected that the query processing was taking
            # too long, as we processed individual results.  This is a different case than dnstable
            # raising Timeout.
            log(DEBUG, "query time exceeded, > " + str(self.max_query_time) + " seconds, aborting")
            yield flexapiserver.response.format_saf_row(cond = 'failed', msg = "Processing timeout").encode('utf-8')
            self.log_lookup(early_exit=True, early_cause="query_time_exceeded")
            return
        except QueryTimeExceededIncomplete:
            self.client.close()
            # Same as QueryTimeExceeded except for the error message
            log(DEBUG, "query time exceeded, > " + str(self.max_query_time) + " seconds, aborting")
            yield failed_incomplete
            self.log_lookup(early_exit=True, early_cause="query_time_exceeded")
            return
        except flexapiserver.response.HTTPResponse as err:
            log_lookup_error(self.env, err.status, self.search_method)
            if sent_headers:
                yield failed_internal
            else:
                self.start_response(err.status, err.headers)
                if err.message is not None:
                    yield err.message
        except Exception as err:
            self.client.close()
            if sent_headers:
                # added 4/9/2021; previously we were not logging such failures
                log(DEBUG, "format_response, general exception", err, traceback.format_exc())
                yield failed_internal
            else:
                status = flexapiserver.response.Raw500Status
                headers = [('Content-Type', 'text/html')]

                tb = traceback.format_exc()
                logid = str(uuid.uuid4())
                log_lookup_error(self.env, status, self.search_method, tb, logid)
                self.start_response(status, headers)
                yield flexapiserver.response.format_http500(logid).encode('utf-8')

    def log_lookup(self, early_exit=False, early_cause=None):
        msg = flexapiserver.util.form_basic_log(self.env, self.search_method)
        msg['count'] = self.row_count
        msg['bytes'] = self.bytes
        msg['time'] = round(self.secs, 3)
        msg['ttfb'] = round(self.secs_to_first, 3) if self.secs_to_first is not None else 0

        if self.id:
            msg['clientId'] = self.id
        if self.swclient:
            msg['swclient'] = self.swclient
        if self.version:
            msg['version'] = self.version
        if early_exit:
            msg['early_exit'] = True

        if early_cause:
            msg['early_cause'] = early_cause

        msg['lookup'] = {
            'server': 'flex',
            'type':   self.what_to_search,
            'value':  self.value,
            'rrtype': self.rrtype,
            'msg':    self.msg if self.msg else "OK"
        }
        if self.otherQueryParamsToLog:
            msg['lookup']['otherParams'] = self.otherQueryParamsToLog

        log(NOTICE, json.dumps(msg))

    def begin_lookup(self):
        time_filter = flexapiserver.filter.TimeFilter()
        time_filter.apply_query_string(self.qs)
        time_filter.apply_policy(self.env.get('flexapiserver.policy', None))

        num_results_override = self.env.get('flexapiserver.res_limit_override')
        if num_results_override:
            try:
                self.limit = min(self.limit, int(num_results_override))
            except:
                log(DEBUG, 'Illegal value of res_limit_override "%s" for a user' % (num_results_override,))

        # if offset > 0 specified and got an num_results_override
        # (presumably for a DCE type of restricted power key), fail
        if self.offset > 0 and num_results_override:
            raise flexapiserver.response.OffsetDisallowed
        # if offset is too high then fail or fail
        if self.offset_max and self.offset > self.offset_max:
            raise flexapiserver.response.OffsetOverMax

        internal_qs = {'search_method': self.search_method,
                       'search_what': self.what_to_search, # note the renaming of the fields
                       'return_what': 'terse', # note the renaming of the fields
                       'value': self.value,
                       'rrtype': self.rrtype,
                       'limit': self.limit}
        if time_filter.time_first_before is not None:
            internal_qs['time_first_before'] = time_filter.time_first_before
        if time_filter.time_first_after is not None:
            internal_qs['time_first_after'] = time_filter.time_first_after
        if time_filter.time_last_before is not None:
            internal_qs['time_last_before'] = time_filter.time_last_before
        if time_filter.time_last_after is not None:
            internal_qs['time_last_after'] = time_filter.time_last_after
        if self.offset is not None:
            internal_qs['offset'] = self.offset
        if self.exclude is not None:
            internal_qs['exclude'] = self.exclude

        if not self.client.send_query(internal_qs):
            raise flexapiserver.response.BackendServerError
        self.t0 = time.time()
        self.resultsIterator = self.client.results_iterator()
