# Copyright (c) 2012-2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import errno
import itertools
import json
import re
import urllib.parse
import ipaddress

from flexapiserver.log import *
import flexapiserver.conf
import flexapiserver.rate
import flexapiserver.response
import flexapiserver.util

# Get our configuration dictionary.  See flexapiserver.conf for the required_config_keys and optional_config_keys.
# get_config() may raise an error, which we let kill the process
conf = flexapiserver.conf.get_config()

# Only accept API keys that match this pattern.
# Previously passed re.IGNORECASE flag, but that is wrong since the API keys become case-sensitive file names.
# NEVER allow a period in these for the same reason.
apikey_pattern = re.compile('^[-a-z0-9]+$')

def get_rate_limiter():
    '''
      If redis is not configured in the flexapiserver.conf file then all quotas are unlimited.
      If redis is configured:
        If there is no quota defined for a particular API key
        or if redis is unavailable upon startup
        or if redis fails after startup and it throws a ConnectionError or TimeoutError
        then quotas are treated as unlimited (and queries work).
        Once redis recovers or a quota is defined for the key then quotas will work normally.
        Other types of failures from redis will return a 500 Server Error.
      '''
    if conf.get('export_installation') is True:
        return None             # unlimited but don't print anything about it
    import redis
    if 'query_limit_redis_url' in conf:
        redis_socket = redis.from_url(conf['query_limit_redis_url'])
        return flexapiserver.rate.RateLimiter(redis_socket, conf['query_limit_period'])
    elif 'query_limit_redis_socket' in conf:
        redis_socket = redis.Redis(unix_socket_path=conf['query_limit_redis_socket'])
        return flexapiserver.rate.RateLimiter(redis_socket, conf['query_limit_period'])
    else:
        print("Not configured to use redis - will use unlimited quotas")
        return None

def log_rate_limit_exception(env):
    msg = flexapiserver.util.form_basic_log(env, 'rate-limit-exception')
    #$$$ For groups: will need to determine if also exceeded the group's quota and report that
    log(NOTICE, json.dumps(msg))

def log_rate_quota_expired(env):
    ''' $$$ for now, the same message as for a rate_limit_exception '''
    msg = flexapiserver.util.form_basic_log(env, 'rate-limit-exception')
    #$$$ For groups: will need to determine if also exceeded the group's quota and report that
    log(NOTICE, json.dumps(msg))

def autoq_env2RateParamsUser(env):
    ''' Generate rate params tuple for a User quota '''
    if env is None:
        return flexapiserver.rate.allNoneAutoQRateParams
    maximum = env.get('flexapiserver.limit')
    if not maximum:
        maximum = conf['query_limit_default']
    return flexapiserver.rate.AutoQRateParams(authtype = env.get('flexapiserver.authtype'),
                                       identifier = env.get('flexapiserver.authuser'),
                                       maximum = maximum,
                                       period = env.get('flexapiserver.period'))

def autoq_env2RateParamsGroup(env):
    ''' Generate rate params tuple for a Group quota '''
    if env is None:
        return flexapiserver.rate.allNoneAutoQRateParams
    return flexapiserver.rate.AutoQRateParams(authtype = 'group',
                                       identifier = env.get('flexapiserver.authgroup'),
                                       maximum = env.get('flexapiserver.group_limit'),
                                       period = env.get('flexapiserver.group_period'))

def blkq_env2RateParamsUser(env):
    ''' Generate rate params tuple for a User quota '''
    if env is None:
        return flexapiserver.rate.allNoneBlkQRateParams
    return flexapiserver.rate.BlkQRateParams(authtype = env.get('flexapiserver.authtype'),
                                      identifier = env['flexapiserver.authuser'],
                                      blkq_split = env.get('flexapiserver.blkq_split'),
                                      blkq_baseline_used = env.get('flexapiserver.blkq_baseline_used'),
                                      blkq_expires = env.get('flexapiserver.blkq_expires'),
                                      burst_size = env.get('flexapiserver.burst_size'),
                                      burst_window = env.get('flexapiserver.burst_window'))

class ratelimitmixin:
    '''
    a base class that requires self.rate to be defined in a subclass
    if either of the base class' methods are called.

    We call the *_limit methods with a list of RateParams namedtuples.

    When creating a namedtuple, we must provide all of the defined elements.
    In our case, some of them will always be None, which is ok.
    '''
    def autoq_get_rate_params(self, env):
        if env is None or self.rate is None:
            return None
        else:
            return [ autoq_env2RateParamsUser(env), autoq_env2RateParamsGroup(env) ]

    def autoq_incr_limit(self, env):
        '''
        This will raise flexapiserver.response.RateLimitExceeded if we exceed the maximum.
        '''
        if env['rp'] is None:
            return
        try:
            self.rate.autoq_incr_limit(env['rp'])
        except flexapiserver.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return flexapiserver.rate.unlimited
        except flexapiserver.rate.LimitExceeded:
            log_rate_limit_exception(env)
            raise flexapiserver.response.RateLimitExceeded
        except Exception:
            log_exception()
            raise flexapiserver.response.InternalServerError

    def autoq_decr_limit(self, env):
        '''
        decrement (undo) a autoq quota unit incr
        '''
        if env['rp'] is None:
            return
        try:
            self.rate.autoq_decr_limit(env['rp'])
        except flexapiserver.rate.RedisConnError:
            return
        except Exception:
            log_exception()

    def blkq_get_rate_params(self, env):
        if env is None or self.rate is None:
            return None
        else:
            return [ blkq_env2RateParamsUser(env) ]

    def blkq_incr_limit(self, env):
        '''
        This will raise flexapiserver.response.RateLimitExceeded if we exceed the maximum.
        '''
        if env['rp'] is None:
            return
        try:
            self.rate.blkq_incr_limit(env['rp'])
        except flexapiserver.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return flexapiserver.rate.unlimited
        except flexapiserver.rate.LimitExceeded:
            log_rate_limit_exception(env)
            raise flexapiserver.response.RateLimitExceeded
        except flexapiserver.rate.BurstLimitExceeded:
            log_rate_limit_exception(env)
            raise flexapiserver.response.BurstRateExceeded
        except flexapiserver.rate.QuotaExpired:
            log_rate_quota_expired(env)
            raise flexapiserver.response.QuotaExpired
        except Exception:
            log_exception()
            raise flexapiserver.response.InternalServerError

    def blkq_decr_limit(self, env):
        '''
        decrement (undo) a blkq quota unit incr
        '''
        if env['rp'] is None:
            return
        try:
            self.rate.blkq_decr_limit(env['rp'])
        except flexapiserver.rate.RedisConnError:
            return
        except Exception:
            log_exception()

empty_policy = object()
empty_acl = object()

def check_acl(env, want_raise=True):
    if not env.get('flexapiserver.policy'):
        return empty_policy
    if not env['flexapiserver.policy'].get('addresses'):
        return empty_acl
    try:
        remote_ip = ipaddress.ip_address(env['REMOTE_ADDR'])
    except:
        # weird -- nginx/uwsgi passed us an invalid IP
        raise  flexapiserver.response.InternalServerError
    for acl in itertools.chain(
            conf.get('admin_addresses', []),
            env['flexapiserver.policy']['addresses'],
            ):
        # check if REMOTE_ADDR is equal or inside the IP or CIDR
        # block specified by acl
        if remote_ip in ipaddress.ip_network(acl):
            return acl
    else:                       # nothing matched
        if want_raise:
            raise flexapiserver.response.RemoteAddressUnauthorized
        return None

def setEnvFromBurstSizeAndWindow(env, burst_size, burst_window):
    """ Sets the burst size and window in env, only if either both are None or both are not None.
        Otherwise sets both to None.
    """
    if (burst_size is None) == (burst_window is None):
        env['flexapiserver.burst_size'] = burst_size
        env['flexapiserver.burst_window'] = burst_window
    else:
        print("Warning burst_size and burst_window are not either both None or both not None")
        env['flexapiserver.burst_size'] = None
        env['flexapiserver.burst_window'] = None

class apikey(ratelimitmixin):
    def __init__(self, app):
        self.app = app
        self.rate = get_rate_limiter()
        # self.last_BadAPIKey is a one-element cache of the last bad key we got.  Due to repeated bad keys,
        # we log a lot.  Don't log if the previous key was the same as this one
        self.last_BadAPIKey = None

    def load_key(self, env):
        '''Given an API key passed in the env array from the query, check its
           format.  Then try to load it from the file system.
           Extract fields from it into the env array.  Raise an error
           if the key is bad.
        '''
        if not 'HTTP_X_API_KEY' in env:
            raise flexapiserver.response.MissingAPIKey

        try:
            api_key = env['HTTP_X_API_KEY']
            if apikey_pattern.match(api_key):
                key_fname = os.path.join(flexapiserver.conf.API_DIR, api_key)
                #$$$ maybe cache based upon the last-mod-time to avoid reopening and parsing it for every query?
                with open(key_fname) as f:
                    keyConfDict = json.load(f)

                env['flexapiserver.authtype'] = 'api'
                env['flexapiserver.authuser'] = keyConfDict['username']
                env['flexapiserver.authgroup'] = keyConfDict.get('group')
                env['flexapiserver.policy'] = keyConfDict.get('policy')
                env['flexapiserver.res_limit_override'] = keyConfDict.get('res_limit_override')
                setEnvFromBurstSizeAndWindow(env, keyConfDict.get('burst_size'), keyConfDict.get('burst_window'))

                # Default the min and max api version in case we don't get values in auth-api files
                env['flexapiserver.min_api_version'] = keyConfDict.get('min_api_version', 2)
                env['flexapiserver.max_api_version'] = keyConfDict.get('max_api_version', 9999)

                # Determine if auto-replenishing or block quota based.
                blkq_split = keyConfDict.get('blkq_split')

                if blkq_split is None:
                    # auto-replenishing quota
                    env['flexapiserver.limit'] = keyConfDict.get('limit')
                    env['flexapiserver.period'] = keyConfDict.get('period')
                    env['flexapiserver.group_limit'] = keyConfDict.get('group_limit')
                    env['flexapiserver.group_period'] = keyConfDict.get('group_period')
                    env['rp'] = self.autoq_get_rate_params(env)
                    env['incr_quota_func'] = self.autoq_incr_limit
                    env['decr_quota_func'] = self.autoq_decr_limit
                else:
                    # block quota
                    env['flexapiserver.blkq_split'] = blkq_split
                    env['flexapiserver.blkq_baseline_used'] = keyConfDict.get('blkq_baseline_used')
                    env['flexapiserver.blkq_expires'] = keyConfDict.get('blkq_expires')
                    env['rp'] = self.blkq_get_rate_params(env)
                    env['incr_quota_func'] = self.blkq_incr_limit
                    env['decr_quota_func'] = self.blkq_decr_limit

            else:
                # Note regarding these print statements: They output to /var/log/syslog (or messages).  We are
                # not using log(NOTICE, ...), which goes to /var/log/dnsdb/dnsdb.notice because that log is
                # mined by the brittle dnsdblog2erp script when it pushes the logs into postgres. log(DEBUG
                # etc are not by default configured by rsyslogd to be actually logged. So a print statement is
                # the only "safe" and default way to get a log message on disk
                if self.last_BadAPIKey == None or self.last_BadAPIKey != api_key:
                    print("BadAPIKey: does not match api key pattern", api_key, "from IP", env['REMOTE_ADDR'])
                self.last_BadAPIKey = api_key
                raise flexapiserver.response.BadAPIKey
        except IOError as e:
            if e.errno == errno.ENOENT:
                # implement a one element memory of the last bad API key logged
                # to reduce duplicate logging
                if self.last_BadAPIKey == None or self.last_BadAPIKey != api_key:
                    print("BadAPIKey: no key registered for key", api_key, "from IP", env['REMOTE_ADDR'])
                self.last_BadAPIKey = api_key
                raise flexapiserver.response.BadAPIKey
            else:
                print("BadAPIKey: other IOError for key", api_key, "from IP", env['REMOTE_ADDR'])
                log_exception()
                raise flexapiserver.response.BadAPIKey
        except Exception as e:
            # since flexapiserver.response.BadAPIKey is an instance, not a pure class, can't catch it directly
            if e is flexapiserver.response.BadAPIKey:
                raise
            print("BadAPIKey: other Exception for key", api_key, "from IP", env['REMOTE_ADDR'])
            log_exception()
            raise flexapiserver.response.BadAPIKey

    def __call__(self, env, start_response):
        try:
            self.load_key(env)
            check_acl(env, want_raise=True) # will raise if our IP is not authorized

            if env['flexapiserver.min_api_version'] > flexapiserver.conf.MY_API_VERSION \
               or env['flexapiserver.max_api_version'] < flexapiserver.conf.MY_API_VERSION:
                raise flexapiserver.response.APIVersionMismatch

            # If limit exceeded, incr_limit will raise a RateLimitExceeded or BurstRateExceeded error
            # and we'll return an HTTP error
            env['incr_quota_func'](env) # ignore returned robj

            # self.app is where the real query processing takes place.
            # It calls api.py:FLEXAPI:__call__() which creates a api.py:Lookup() object.
            return self.app(env, start_response)

        except flexapiserver.response.HTTPResponse as res:
            # handle any kind of quota limit specially -- don't give back quota units for them!
            if res.status != flexapiserver.response.Raw429Status \
               and res.status != flexapiserver.response.Raw401Status \
               and res.status != flexapiserver.response.BadAPIKey.status:
                env['decr_quota_func'](env)

            start_response(res.status, res.headers)
            return [res.message]

        except Exception as e:
            log_exception()
            res = flexapiserver.response.InternalServerError
            start_response(res.status, res.headers)
            return [res.message]
